#include <cstdlib>
#include <cstdio>
#include <algorithm>
#include <assert.h>
#include <math.h>
#include <string>
#include <chrono>
#include <iostream>
#include <algorithm>
#include <vector>
#include <unistd.h>
#include <iostream>
#include <fstream>
#include <cstring>
#include <cstdint>
#include <climits>
#include <array>

#ifdef OPENACC
#include <accelmath.h>
#include <openacc.h>
#ifdef CUDA
#include <cuda.h>
#include <cuda_runtime.h>
#endif
#else
#include <omp.h>
#endif

#define MIN(a, b) ((a<b)?a:b)

const long int G=27;
const unsigned int GL=8000000;
const int N=0;
const long int K=G;
const unsigned int max_loop=30;
const int cuba=8;
const int Nbin=128;
const unsigned int BLt=GL/Nbin;
const unsigned int GL1=BLt-1;
int STEP;

struct Particle
{
  float        rx;
  float        ry;
  float        rz;
  float        ct;
  float        vx;
  float        vy;
  float        vz;
  float        en;
  float        de;
  int          ix;
  int          jy;
  int          kz;
  int          ip;
  float        wt;
  unsigned int rs;
  int          ir;
  int id;
};

Particle particles[GL];// __attribute__((aligned(64)));        // Nth or GL?

int ind01[Nbin][BLt];// __attribute__((aligned(64)));
int ind23[Nbin][BLt];// __attribute__((aligned(64)));

long int MAX_ELEMENT;
unsigned int SHIFT;
unsigned int POSITION0;
unsigned int POSITION1;
unsigned int POSITION2;
unsigned int POSITION23;
unsigned int POSITION3;

double DGam;
double Gam;
double NaN;
unsigned int Ntop=0;
unsigned int LIFE=0;
double SumDGam=0.;
double GAMMA=0.;
double NoNew=0.;
float cgam=5.0f;          
float dcgam=cgam+cgam;    
const int cubn=cuba+cuba; 
const int cub2=cubn*cubn; 
const int Ngn=GL/cubn;   
const int cub3=cub2*cubn; 

const double fuse=0.70;  
double oldL=1.;
double oldW=1.;
double oldT=0.;

int mini[Nbin];
int count01[Nbin];
int count23[Nbin];
int count0[Nbin];
int count1[Nbin];
int count2[Nbin];
int pos2[Nbin];
int count3[Nbin];
int ii1[Nbin];
int ii3[Nbin];
int ii23[Nbin];

int init[Nbin];
int fin[Nbin];

int init23[Nbin];
int fin23[Nbin];

int init3[Nbin];
int fin3[Nbin];

Particle arr1[GL] __attribute__((aligned(64)));  
Particle arr2[GL] __attribute__((aligned(64)));  
Particle arr3[GL] __attribute__((aligned(64)));  
unsigned int sizep=sizeof(Particle);

int pointer1[Nbin];
int pointer2[Nbin];
int pointer3[Nbin];



//
constexpr double DT=1.0e-3;//this is 1 microsecond
double TimeLimit=DT;
bool AllNeutronsAreAtTimeLimit=false;
int NumOf=0;
//consider each voxel is a cube with rib=1 mm:
//U density=18.95 gr/cm^3=0.01895 gr/mm^3
constexpr double U_density=18.95/1000;
//Avogadro number:
constexpr double NA = 6.022e23;//mole^(-1).  
//molar mass of U238, gr/mole:
constexpr double Y_molar_mass=238.03;//
//initial number of U atoms in 1 voxel:
constexpr double N0=U_density/Y_molar_mass*NA;
//number of burnout U atoms in voxels:
double Uburnout[cub3]{0.0};
//number of neutrons in each voxel:
int Length[cub3]{0};
//offsets of indices of different voxels in array MIndex:
int Begin[cub3]{0};
//for filling in MIndex in injector()
int Counter[cub3]{0};
//indices of neutrons of different voxels:
int MIndex[GL]{0};
//instead of mmn
//This helps to store which neutrons to stick together and which not.
int MMN[GL]{0};
//
#ifdef OPENACC
#pragma acc routine seq
#endif
int Get1DVoxelIndex(const Particle & p){return (p.ix+cuba+cubn*(p.jy+cuba+cubn*(p.kz+cuba)));};
//

#ifndef OPENACC
struct float3
{
  float x;
  float y;
  float z;
};

float3 operator/(const float3 v, const float a)
{
  float3 result{v.x/a, v.y/a, v.z/a};
  return result;
}
float3 operator*(const float a, const float3 v)
{
  float3 result{v.x*a, v.y*a, v.z*a};
  return result;
}
float operator*(const float3 v0, const float3 v1)
{
  return v0.x*v1.x+v0.y*v1.y+v0.z*v1.z;
}
#else
float operator*(const float3 v0, const float3 v1)
{
  return v0.x*v1.x+v0.y*v1.y+v0.z*v1.z;
}
#endif

void operator*=(float3 & f,const float a) {f.x*=a, f.y*=a, f.z*=a;}
void operator/=(float3 & f,const float a) {f.x/=a, f.y/=a, f.z/=a;}
void operator+=(float3 & f, const float3 & p) {f.x+=p.x, f.y+=p.y, f.z+=p.z;}
void operator-=(float3 & f, const float3 & p) {f.x-=p.x, f.y-=p.y, f.z-=p.z;}


float3 operator*(const float3 v, const float a)
{
  float3 result{v.x*a, v.y*a, v.z*a};
  return result;
}

float3 operator+(const float3 v0, const float3 v1)
{
  float3 result{v0.x+v1.x, v0.y+v1.y, v0.z+v1.z};
  return result;
}

float3 operator-(const float3 v0, const float3 v1)
{
  float3 result{v0.x-v1.x, v0.y-v1.y, v0.z-v1.z};
  return result;
}

float3 cross(float3 v0, float3 v1)
{
  float3 rt;
  rt.x = v0.y*v1.z-v0.z*v1.y;
  rt.y = v0.z*v1.x-v0.x*v1.z;
  rt.z = v0.x*v1.y-v0.y*v1.x;
  return rt;
}
float mod2(float3 vector)
{
  return vector*vector;
}
float3 normalize(float3 vector)
{
  float dist = sqrtf(vector*vector);
  vector/=dist;
  return vector;
}

unsigned int Rand32(unsigned int xn)
{
  u_quad_t a=0x5DEECE66D;
  u_quad_t c=0xB;
  return (unsigned int)((a*xn+c) & 0xFFFFFFFF);
}

double rndv(unsigned int xn)
{
  return (double) xn / (double) 0x100000000LL;
}

inline float RND_Gaussian_theta(float disp, unsigned int xn) // Pseudo Gauss 
{
  const float dtheta0=0.01f; // This is an external parameter !!!!
  xn=Rand32(xn);
  float p1=float(1.-2*rndv(xn));
  float p12=p1*p1;
  float lips=sqrt(1.f-p1*p1)-1.E-5;
  xn=Rand32(xn);
  float p2=lips*float(1.-2*rndv(xn));
  float p=p12+p2*p2;
  if(p>=1.) p=.9999999;
  return fabs(dtheta0*p1*sqrtf(-2*disp*logf(p)/p));
}

bool compar(std::pair<int,int> i, std::pair <int,int> j) {return (int)(i.second) > (int)(j.second);}

void propagator()
{
  /*const */float lambda=1.0f; // EM radiation length
  const float lambva=.01f; // Nuclear interaction length / velocity
  const float ag=1.0f;
  const float da=ag*1e-4;
  const float rho0=1.0f;
//!!!NOTE!!!
//#pragma omp parallel for simd here gives different results
//compared to if to remove all OpenMP pragmas before this for-loop:
//1) with #pragma omp parallel for simd
//158   504    357947    29   1
//Uburnout: 278952
//2) with #pragma omp parallel for schedule(dynamic) private(lambda)
//158   526    400842    29   1
//Uburnout: 384419
//SO, NEVER USE
//#pragma omp parallel for simd
//because it gives different to non-parallel code results.
//REMOVE simd
//!!!VERY IMPORTANT!!!
//
//!!!  
//But adding schedule(dynamic) clause increases execution time to more than 130 seconds (max_loop=5000)
//But removing schedule(dynamic) clause reduces execution time to some 80 seconds
//!!!
#ifdef OPENACC
#pragma acc parallel loop gang vector copy(LIFE) present(particles,Uburnout) copy(lambda,lambva,ag,da,rho0,N0) reduction(+:)
#else
#pragma omp parallel for /*simd*/ /*schedule(dynamic)*/ private(lambda)
#pragma distribute_point
#endif
  for(int i=0; i<LIFE; ++i) if(particles[i].ir>0 &&//Exclude compressed particles
                               //Time limit:
                               fabs(particles[i].ct-TimeLimit) > 1.0e-3*DT) 
  {
    float l1x;
    float l1y;
    float l1z;
    if(particles[i].vx>=0.0f) l1x=((particles[i].ix+1)*ag-particles[i].rx)/particles[i].vx+da;
    else                      l1x=(particles[i].ix*ag-particles[i].rx)/particles[i].vx+da;
    if(particles[i].vy>=0.0f) l1y=((particles[i].jy+1)*ag-particles[i].ry)/particles[i].vy+da;
    else                      l1y=(particles[i].jy*ag-particles[i].ry)/particles[i].vy+da;
    if(particles[i].vz>=0.0f) l1z=((particles[i].kz+1)*ag-particles[i].rz)/particles[i].vz+da;
    else                      l1z=(particles[i].kz*ag-particles[i].rz)/particles[i].vz+da;
    ///if(isnan(l1x) || isnan(l1y) || isnan(l1z)) std::cout<<"v="<<particles[i].vx<<" "<<particles[i].vy<<" "
    ///                                                    <<particles[i].vz<<std::endl;
//Burnout:
    //voxel number:
    int vn_old=Get1DVoxelIndex(particles[i]);
    
    lambda=1.0f*N0/(N0-Uburnout[vn_old]);
//      
    int m=40;
    //was an ERROR!
    //float l1=m*MIN(MIN(l1x, l1y), MIN(l1y, l1z));  // THIS IS WRONG!!!
    //CORRECT VARIANT WITHOUT m:
    float l1=MIN(MIN(l1x, l1y), MIN(l1y, l1z));
    float lamb=m*lambda;             // EM default
    if(particles[i].ip==3) lamb=m*lambva*particles[i].en;
    particles[i].rs=Rand32(particles[i].rs);
    double f2=rndv(particles[i].rs);
    float l2=fabs(lamb*log(f2));
    particles[i].rs=Rand32(particles[i].rs);
    double f3=rndv(particles[i].rs);
    if(particles[i].ip==3) lamb/=5.;
    float l3=fabs(lamb*log(f3));
    int irc;
    float l23;
    if(l3<l2)
    {
      l23=l3;
      irc=3;
    }
    else
    {
      irc=2;
      l23=l2;
    }
    float l=l1;
    if(l23<l1) l=l23;
    else       irc=1;
    ///if(isnan(l)) std::cout<<"*** l=nan"<<" irc="<<irc<<" l1="<<l1<<" l23="<<l23<<std::endl;
    //Orts: e1 e2
    float3 e1;
    if(particles[i].vx<particles[i].vy)
    {
      if(particles[i].vx<particles[i].vz) e1={0., particles[i].vz, -particles[i].vy};
      else                                e1={particles[i].vy, -particles[i].vx, 0.};
    }
    else
    {
      if(particles[i].vy<particles[i].vz) e1={particles[i].vz, 0., -particles[i].vx};
      else                                e1={particles[i].vy, -particles[i].vx, 0.};
    }
    float3 e2={particles[i].vy*e1.z-particles[i].vz*e1.y,
               particles[i].vz*e1.x-particles[i].vx*e1.z,
               particles[i].vx*e1.y-particles[i].vy*e1.x};
    e1=normalize(e1);
    e2=normalize(e2);
    particles[i].rs=Rand32(particles[i].rs);
    double phi=rndv(particles[i].rs)*2*M_PI;
    //угол многократного рассеяния
    const float dtheta0=0.01f; // This is an external parameter !!!!
    particles[i].rs=Rand32(particles[i].rs);
    float p1=float(1.-2*rndv(particles[i].rs));
    float p12=p1*p1;
    float lips=sqrt(1.f-p1*p1)-1.E-5;
    particles[i].rs=Rand32(particles[i].rs);
    float p2=lips*float(1.-2*rndv(particles[i].rs));
    float p=p12+p2*p2;
    if(p>=1.) p=.9999999;
    float theta=fabs(dtheta0*p1*sqrtf(-2*fabs(l)*logf(p)/p));
    if(theta>1.) theta=1.;
    float cost=cos(theta);
    float sint=0.;
    if(cost<1.f && cost>-1.f) sint=sqrtf(1.f-cost*cost);
    float3 vl={particles[i].vx*cost, particles[i].vy*cost, particles[i].vz*cost};
    float3 new_vel=vl+e1*(sint*sin(phi))+e2*(sint*cos(phi));
    ///if(isnan(new_vel.x)||isnan(new_vel.y)||isnan(new_vel.z))
    ///  std::cout<<"***pv="<<particles[i].vx<<" "<<particles[i].vy<<" "<<particles[i].vz<<" phi="<<phi
    ///           <<" ct="<<cost<<std::endl;
    new_vel=normalize(new_vel);
    //скорость
    particles[i].vx=new_vel.x;
    particles[i].vy=new_vel.y;
    particles[i].vz=new_vel.z;
    float dl=fabs(l);

    /*
    if(particles[i].id==9)
      std::cout<<"### l1x="<<l1x<<" l1y="<<l1y<<" l1z="<<l1z<<" l2="<<l2<<" l3="<<l3
               <<" l="<<l<<" dl="<<dl<<std::endl;
    */
    
////Time limit://////////////////////////////////////////////////////////////
    //velocity:
    const double v=sqrt(particles[i].vx*particles[i].vx+
                        particles[i].vy*particles[i].vy+
                        particles[i].vz*particles[i].vz)*
                   particles[i].en;

    //std::cout<<"dl/v="<<dl/v<<std::endl;
    //if(particles[i].ct/DT > TimeLimit/DT)
    //  std::cout<<"ct="<<particles[i].ct/DT<<" TM="<<TimeLimit/DT<<std::endl;

    const double prev_ct=particles[i].ct;
    if((dl+da)/v < 0.0)
      printf("***ERROR: dl=%f\n",dl);
    
    if(particles[i].ct + (dl+da)/v > TimeLimit)
    {
      //std::cout<<"HERE"<<std::endl;
      dl=(TimeLimit-particles[i].ct)*v;
      particles[i].ct+=dl/v;
    }
    else
      particles[i].ct+=(dl+da)/v;

    if(particles[i].ct < prev_ct)
    {
      printf("***ERROR: prev_ct=%f particles[%d].ct=%f\n",
             prev_ct,i,particles[i].ct);
      ///exit(0);
    }
    if(particles[i].ct > TimeLimit + 1.0e-3*DT)
    {
      printf("***ERROR: ct=%f > TimeLimit=%f\n",particles[i].ct,TimeLimit);
      ///exit(0);
    }
//////////////////////////////////////////////////////////////////////////////
    
    //was an ERROR:
    //particles[i].rx+=particles[i].vx*dl+da;
    //particles[i].ry+=particles[i].vy*dl+da;
    //particles[i].rz+=particles[i].vz*dl+da;
    //here an error was:
    //just imagine that the particle velocity is in the negative direction of Ox or Oy or Oz.
    //make_particle() makes new particles in coordinate point(0,0,0) with
    //velocity direction vdir=(1.0f/sqrtf(3.0),1.0f/sqrtf(2.0),1.0f/sqrtf(6.0)).
    //SDI made random make_particle() with random initial velocity=+=vdir.
    //The particle is standing on the edge of the voxel (cube cell),
    //l1=0 and the progrma chooses ir=1 (free path). Since dl=0, the new coordinates of the particle are:
    //particles[i].rx+=particles[i].vx*dl+da=0+da=da;
    //particles[i].ry+=particles[i].vy*dl+da=0+da=da;
    //particles[i].rz+=particles[i].vz*dl+da=0+da=da;
    //da=1.0e-4 - very small.
    //The velocity is directed opposite to the displacement of the particle. That is why the particle starts
    //moving back. But since l1 is very small <= da, the particle doesn't manage to cross the edge of the voxel.
    //It moves to (da,da,da) nad back again and again.
    //THIS IS THE ERROR CYCLING (zaciklivaniye).
    //TO CORRECT THIS, WE SHOULD USE the direction of the particle velocity in the safety da:
    //particles[i].rx+=particles[i].vx*(dl+da);
    //particles[i].ry+=particles[i].vy*(dl+da);
    //particles[i].rz+=particles[i].vz*(dl+da);
    //THIS IS RIGHT!
    //CORRECT VARIANT:
    particles[i].rx+=particles[i].vx*(dl+da);
    particles[i].ry+=particles[i].vy*(dl+da);
    particles[i].rz+=particles[i].vz*(dl+da);


    ///std::cout<<"ct="<<particles[i].ct<<"|"<<TimeLimit<<std::endl;
    
    
    /*
    if(particles[i].id == 9)
    {
      std::cout<<"###particles["<<i<<"].id="<<particles[i].id
               <<" ct="<<particles[i].ct
               <<" prev_ct="<<prev_ct
               <<" ir="<<irc
               <<" dl="<<dl
               <<" da="<<da
               <<" ag="<<ag
               <<" l1x="<<l1x<<" l1y="<<l1y<<" l1z="<<l1z
               <<" l2="<<l2<<" l3="<<l3
               <<" v="<<v
               <<" (dl+da)/v="<<(dl+da)/v
               <<" TimeLimit="<<TimeLimit
               <<std::endl;
    }
    */

    ///if(isnan(particles[i].rx) || isnan(particles[i].ry) || isnan(particles[i].rz))  
    ///  std::cout<<"i="<<i<<" dl="<<dl<<" v="<<particles[i].vx<<" "<<particles[i].vy<<" "<<particles[i].vz<<std::endl;
    particles[i].ix=std::floor(particles[i].rx/ag);
    particles[i].jy=std::floor(particles[i].ry/ag);
    particles[i].kz=std::floor(particles[i].rz/ag);

//Burnout:    
    //voxel number:
    int vn_new=Get1DVoxelIndex(particles[i]);
    //number of burnout U atoms in voxel:
#ifdef OPENACC
  #pragma acc atomic update  
#else
  #pragma omp atomic update
#endif
    Uburnout[vn_new] += particles[i].wt;
//End of burnout.    
    
    bool out=(particles[i].ix>=cuba || particles[i].jy>=cuba || particles[i].kz>=cuba ||
              particles[i].ix<-cuba || particles[i].jy<-cuba || particles[i].kz<-cuba);
    float loss=0.;
    if(out)                       // kill outside Crit Cube
    {
      particles[i].de=particles[i].en;
      particles[i].en=0.;
      particles[i].ir=0;
    }
    else if(particles[i].ip==3 && irc==3)
    {
      particles[i].ir=3;
      particles[i].de+=particles[i].en-dcgam; // E-conserv Corr
    }
    else if(particles[i].ip!=0 && particles[i].ip!=3)
    {
      loss=rho0*dl;
      if(loss>=particles[i].en || particles[i].en<=1.) // kill outside Crit Cube
      {
        particles[i].de=particles[i].en;
        particles[i].en=0.;
        particles[i].ir=0;
      }
      else
      {
        particles[i].en-=loss;
        particles[i].de=loss;
        particles[i].ir=irc;
      }
    }
  }
} // End of Propagator

void make_particle()
{
#ifdef OPENACC
  #pragma acc parallel num_gangs(1) vector_length(1) present(particles) copy(LIFE,MAX_ELEMENT)
  {
#endif    
  particles[LIFE].id=MAX_ELEMENT;
  particles[LIFE].wt=1.f;
  particles[LIFE].rs=++MAX_ELEMENT;
  particles[LIFE].rx=0.0f;
  particles[LIFE].ry=0.0f;
  particles[LIFE].rz=0.0f;
  particles[LIFE].ct=0.0f;
  particles[LIFE].en=G;
  particles[LIFE].de=0.0f;
  particles[LIFE].vx=1.0f/sqrtf(3.0);
  particles[LIFE].vy=1.0f/sqrtf(2.0);
  particles[LIFE].vz=1.0f/sqrtf(6.0);
  particles[LIFE].ix=0;
  particles[LIFE].jy=0;
  particles[LIFE].kz=0;
  particles[LIFE].ir=1;
  particles[LIFE].ip=3;
  ++LIFE;
#ifdef OPENACC
  }
#endif
} // End of make_particle

void compressor()
{
//////////////////////////////////////////////////////////////////////////////
//Check if all the particles are in proper time = TimeLimit.
//If even a single particle has ct < TimeLimit, we need to wait
//until all the particles have ct=TimeLimit.
  AllNeutronsAreAtTimeLimit=true;
  NumOf=0;
#ifdef OPENACC
  #pragma acc parallel loop gang vector copy(AllNeutronsAreAtTimeLimit,TimeLimit,DT) reduction(+:NumOf)
#else
  #pragma omp parallel for reduction(+:NumOf)
#endif
  for(int i=0; i<LIFE; ++i)
  {
    if(particles[i].ir>0)
    {
      ///if(fabs(particles[i].ct-TimeLimit) > 1.0e-3*DT)
      ///  particles[i].ct=TimeLimit;
      if(fabs(particles[i].ct-TimeLimit) > 1.0e-3*DT)
      {
        /*
        std::cout<<"particles["<<i<<"].ct="<<particles[i].ct
                 <<" TM="<<TimeLimit
                 <<" r: "<<particles[i].rx<<" "
                 <<particles[i].ry<<" "
                 <<particles[i].rz
                 <<" MAX_ELEMENT="<<MAX_ELEMENT
                 <<" id="<<particles[i].id
                 <<" ir="<<particles[i].ir
                 <<std::endl;
        */
    #ifdef OPENACC
      #pragma acc atomic write
    #else
      #pragma omp atomic write
    #endif
        AllNeutronsAreAtTimeLimit=false;
        ++NumOf;
        //break;
      }
      /*
      if(particles[i].ct > TimeLimit+1.0e-3*DT)
        std::cout<<"***ERROR: particles["<<i<<"].ct="<<particles[i].ct
                 <<"> TimeLimit="<<TimeLimit
                 <<std::endl;
      */
    }
  }
//When All the particle have ct=TimeLimit, we may update TimeLimit:
  if(AllNeutronsAreAtTimeLimit) TimeLimit += DT;
//////////////////////////////////////////////////////////////////////////////  
  
#ifdef OPENACC
  #pragma acc parallel loop gang vector reduction(+:SumDGam) present(particles)
#else
  #pragma omp parallel for simd reduction(+:SumDGam)
  #pragma distribute_point
#endif
 for(int i=0; i<LIFE; ++i)
  {
    SumDGam += (double)particles[i].de*(double)particles[i].wt;
    particles[i].de=0.;
  }
  int dL=LIFE/Nbin;
  int DL=dL+1;
  int n;
  if(LIFE%Nbin==0) n=Nbin;
  else     n=Nbin*DL-LIFE;
  
  POSITION0=POSITION1=POSITION2=POSITION3=POSITION23=0;

#ifdef OPENACC
#pragma acc parallel loop gang vector copy(GL1,dL,DL,n,count01,count23,count0,count1,count2,count3,init,fin)
#else
#pragma omp parallel for simd
#pragma distribute_point
#endif
  for(int b=0; b<Nbin; ++b)    
  {
    count01[b]=GL1;
    count23[b]=0;
    count0[b]=GL1;
    count1[b]=0;
    count2[b]=GL1;
    count3[b]=0;
    if(b<n)
    {
      init[b]=b*dL;
      fin[b]=(b+1)*dL;
    }
    else if(b==n)
    {
      init[b]=n*dL;
      fin[b]=n*dL+DL;
    }
    else if(b>n)
    {
      init[b]=n*dL+DL*(b-n);
      fin[b]=n*dL+DL*(b-n+1);
    }
  }
//сортировка
//count23
#ifdef OPENACC
  #pragma acc parallel loop gang vector copy(count01,count23,init,fin) present(particles,ind23)
#else
  #pragma omp parallel for
#endif
  for(int b=0; b<Nbin; ++b)    
  {
    for(int i=init[b]; i<fin[b]; ++i)
    {
      if(particles[i].ir<2) ind23[b][count01[b]--]=i;
      else                  ind23[b][count23[b]++]=i;
    }
  }
#ifdef OPENACC
  #pragma acc parallel loop gang copy(count0,count1,count2,count3,count01,count23,mini,ii23,init,fin) present(particles,ind01,ind23)
#else
  #pragma omp parallel for
  #pragma distribute_point
#endif
  for(int b=0; b<Nbin; ++b)    
  {
// ============ Separate 0,1 from 2,3==================    
    ii23[b]=count23[b]-1;
    mini[b]=GL1-count01[b];
    if(count23[b]<mini[b]) mini[b]=count23[b];
    int js=0;
#ifdef OPENACC
  #pragma acc loop vector reduction(+:js)
#else
  #pragma omp simd reduction(+:js)
#endif
    for(int j=0; j<mini[b]; ++j) js+=int(ind23[b][ii23[b]-j]>ind23[b][GL1-j]);
#ifdef OPENACC
  #pragma acc loop vector
#else
  #pragma omp simd
#endif
    for(int j=0; j<js; ++j) std::swap(particles[ind23[b][ii23[b]-j]],particles[ind23[b][GL1-j]]);
    for(int i=init[b]; i<fin[b]; ++i)
    {
      if     (particles[i].ir==0) ind01[b][count0[b]--]=i;
      else if(particles[i].ir==1) ind01[b][count1[b]++]=i;
      else if(particles[i].ir==2) ind23[b][count2[b]--]=i;
      else                        ind23[b][count3[b]++]=i;
    }
  }
    
#ifdef OPENACC
  #pragma acc parallel loop gang copy(count0,count1,count2,count3,mini,ii1,ii3,GL1) present(particles,ind01,ind23)
#else
  #pragma omp parallel for
  #pragma distribute_point
#endif
  for(int b=0; b<Nbin; ++b)    
  {
// ============ Sort 01 ==================
    ii1[b]=count1[b]-1;
    mini[b]=GL1-count0[b];
    if(count1[b]<mini[b]) mini[b]=count1[b];
    int js=0;
#ifdef OPENACC
  #pragma acc loop vector reduction(+:js)
#else
  #pragma omp simd reduction(+:js)
#endif
    for(int j=0; j<mini[b]; ++j) js+=int(ind01[b][ii1[b]-j]>ind01[b][GL1-j]);
#ifdef OPENACC
  #pragma acc loop vector
#else
  #pragma omp simd
#endif
    for(int j=0; j<js; ++j) std::swap(particles[ind01[b][ii1[b]-j]],particles[ind01[b][GL1-j]]);
// ============ Sort 23 ==================      
    ii3[b]=count3[b]-1;
    mini[b]=GL1-count2[b];
    if(count3[b]<mini[b]) mini[b]=count3[b];
    js=0;
#ifdef OPENACC
  #pragma acc loop vector reduction(+:js)
#else
  #pragma omp simd reduction(+:js)
#endif
    for(int j=0; j<mini[b]; ++j) js+=int(ind23[b][ii3[b]-j]>ind23[b][GL1-j]);
#ifdef OPENACC
  #pragma acc loop vector
#else
  #pragma omp simd
#endif
    for(int j=0; j<js; ++j) std::swap(particles[ind23[b][ii3[b]-j]],particles[ind23[b][GL1-j]]);
  }
  
// Reorder the pointers limits
#ifdef OPENACC
  #pragma acc parallel loop gang vector reduction(+:POSITION0,POSITION1,POSITION2,POSITION3,POSITION23)
#else
  #pragma omp parallel for reduction(+:POSITION0,POSITION1,POSITION2,POSITION3,POSITION23)
#endif
  for(int b=0; b<Nbin; ++b)
  {
    count0[b]=GL1-count0[b];
    count2[b]=GL1-count2[b];
    POSITION0+=count0[b];
    POSITION1+=count1[b];
    POSITION2+=count2[b];
    POSITION3+=count3[b];
    POSITION23+=count23[b];
  }
  LIFE-=POSITION0;
  SHIFT=LIFE;
  LIFE+=POSITION23;

//здесь должно идти слияние мини ящиков в ящики для 0, 1, 2, 3, удаление 0, и перекладывание 3, 2, 1 в исходный ящик
  pointer1[0]=pointer2[0]=pointer3[0]=0;
#ifdef OPENACC
#pragma acc serial loop /*num_gangs(1) vector_length(1)*/ copy(pointer1,pointer2,pointer3)
#endif
  for(int b=0; b<Nbin-1; ++b)
  {
    pointer1[b+1]=pointer1[b]+count1[b];
    pointer2[b+1]=pointer2[b]+count2[b];
    pointer3[b+1]=pointer3[b]+count3[b];
  }

  //DO NOT parallelize or vectorize - undefined behavior
  for(int b=0; b<Nbin; ++b)
  {
#ifdef OPENACC
#pragma acc host_data use_device(particles,arr1,arr2,arr3)
    {
      cudaMemcpy(&arr1[pointer1[b]],&particles[init[b]+count3[b]+count2[b]],count1[b]*sizep,cudaMemcpyDeviceToDevice);
      cudaMemcpy(&arr2[pointer2[b]],&particles[init[b]+count3[b]],count2[b]*sizep,cudaMemcpyDeviceToDevice);
      cudaMemcpy(&arr3[pointer3[b]],&particles[init[b]],count3[b]*sizep,cudaMemcpyDeviceToDevice);
    }
#else
    memcpy(&arr1[pointer1[b]],&particles[init[b]+count3[b]+count2[b]],count1[b]*sizep);
    memcpy(&arr2[pointer2[b]],&particles[init[b]+count3[b]],count2[b]*sizep);
    memcpy(&arr3[pointer3[b]],&particles[init[b]],count3[b]*sizep);
#endif
  }

  // слияние ящиков для 1, 2, 3 в массив particles
#ifdef OPENACC
#pragma acc host_data use_device(particles,arr1,arr2,arr3)
  {
    cudaMemcpy(&particles[0],&arr3[0],POSITION3*sizep,cudaMemcpyDeviceToDevice);
    cudaMemcpy(&particles[POSITION3],&arr2[0],POSITION2*sizep,cudaMemcpyDeviceToDevice);
    cudaMemcpy(&particles[POSITION23],&arr1[0],POSITION1*sizep,cudaMemcpyDeviceToDevice);
  }
#else
  memcpy(&particles[0],&arr3[0],POSITION3*sizep);
  memcpy(&particles[POSITION3],&arr2[0],POSITION2*sizep);
  memcpy(&particles[POSITION23],&arr1[0],POSITION1*sizep);
#endif
  
#ifdef OPENACC
  #pragma acc parallel loop gang vector copy(SHIFT,MAX_ELEMENT) present(particles)
#else
  #pragma omp parallel for simd
#endif
  for(int i=0; i<POSITION3; ++i)
  {
    int i2=i+i;
    particles[i].rs=MAX_ELEMENT+i2;
    particles[i+SHIFT].rx=particles[i].rx;
    particles[i+SHIFT].ry=particles[i].ry;
    particles[i+SHIFT].rz=particles[i].rz;
    particles[i+SHIFT].ct=0.0;
    particles[i+SHIFT].ix=particles[i].ix;
    particles[i+SHIFT].jy=particles[i].jy;
    particles[i+SHIFT].kz=particles[i].kz;
    particles[i+SHIFT].wt=particles[i].wt;
    particles[i+SHIFT].rs=MAX_ELEMENT+i2+1;
    particles[i+SHIFT].id=MAX_ELEMENT+i2;
    if(particles[i].ip==0)      // gamma
    {
      particles[i].ip=1;
      particles[i+SHIFT].ip=2;
    }
    else if(particles[i].ip==1) // electron
    {
      particles[i].ip=0;
      particles[i+SHIFT].ip=1;
    }
    else if(particles[i].ip==2) // positron
    {
      particles[i].ip=0;
      particles[i+SHIFT].ip=0;
    }
    else if(particles[i].ip==3) // neutrons
    {
      particles[i].ip=3;
      particles[i+SHIFT].ip=3;
    }
  }
  MAX_ELEMENT+=POSITION3+POSITION3;

#ifdef OPENACC
  #pragma acc parallel loop gang vector copy(POSITION3,POSITION23,SHIFT,MAX_ELEMENT) present(particles)
#else
  #pragma omp parallel for simd
#endif
  for(int i=POSITION3; i<POSITION23; ++i)
  {
    particles[i+SHIFT].rx=particles[i].rx;
    particles[i+SHIFT].ry=particles[i].ry;
    particles[i+SHIFT].rz=particles[i].rz;
    particles[i+SHIFT].ct=0.0;
    particles[i+SHIFT].ix=particles[i].ix;
    particles[i+SHIFT].jy=particles[i].jy;
    particles[i+SHIFT].kz=particles[i].kz;
    particles[i+SHIFT].wt=particles[i].wt;
    particles[i+SHIFT].rs=MAX_ELEMENT+i;
    particles[i+SHIFT].id=MAX_ELEMENT+i;
    if(particles[i].ip==0) particles[i+SHIFT].ip=1;
    else                   particles[i+SHIFT].ip=0;
  }
  MAX_ELEMENT+=POSITION23-POSITION3;
} // End of Compressor

void reactor()
{
  float Dan=0.1f;
//here the execution time does not depend on schedule(dynamic)
//But simd to other than in sequential code results.
//That is why removed simd here.
#ifdef OPENACC
  #pragma acc parallel loop gang vector copy(Dan,cgam)
#else
  #pragma omp parallel for /*simd*/ schedule(dynamic)
  #pragma distribute_point
#endif
  for(int i=0; i<POSITION23; ++i)
  {
    float gamma=particles[i].en;
	  float gamma1=cgam;
    float gamma2=cgam;
    if(particles[i].ip!=3 || particles[i].ir!=3) // only (n,2n) doesn't conserves energy
    {
      if(gamma>2.0f)
      {
        particles[i].rs=Rand32(particles[i].rs);
        float part=rndv(particles[i].rs);
        gamma2=gamma*part;
        gamma1=gamma-gamma2;
      }
      else
      {
	      gamma1=gamma/2;
        gamma2=gamma1;    
      }
    }
    float xy=particles[i].vx*particles[i].vy;
    float xz=particles[i].vx*particles[i].vz;
    float yz=particles[i].vy*particles[i].vz;
    float x2=particles[i].vx*particles[i].vx;
    float y2=particles[i].vy*particles[i].vy;
    float z2=particles[i].vz*particles[i].vz;
	  float3 e1;
    float3 e2;
	  if(particles[i].vx<particles[i].vy)
    {
      if(particles[i].vx<particles[i].vz)
      {
        e1={0., particles[i].vz, -particles[i].vy};
        e2={y2+z2, -xy, -xz};
      }
      else
      {
        e1={particles[i].vy, -particles[i].vx, 0.};
        e2={-xz, -yz, y2+x2};
      }
    }
    else
    {
	    if(particles[i].vy<particles[i].vz)
      {
        e1={particles[i].vz, 0., -particles[i].vx};
        e2={xy, -x2-z2, yz};
      }
	    else
      {
        e1={particles[i].vy, -particles[i].vx, 0.};
        e2={-xz, -yz, y2+x2};
      }
    }
	  e1=normalize(e1);
	  e2=normalize(e2);
    particles[i].rs=Rand32(particles[i].rs);
	  float phi=rndv(particles[i].rs)*2*M_PI;
    // first particle
    float cost;
    if(particles[i].ip==3)
    {
      particles[i].rs=Rand32(particles[i].rs);
      cost=1.f-2*rndv(particles[i].rs); // Isotropic nA interaction
    }
    else              // Fixed EM scattering angle
    {
	    float Theta=1.;
      if(gamma>Dan) Theta=Dan/gamma;
      cost=cos(Theta);
    }
    float sint=0.;
    if(cost<1.f && cost>-1.f) sint=sqrtf(1.f-cost*cost);
    float sf=sinf(phi);
    float cf=cosf(phi);
    float ss=sint*sf;
    float sc=sint*cf;

    float3 vc{particles[i].vx*cost, particles[i].vy*cost, particles[i].vz*cost};
    float3 e12{e1.x*ss+e2.x*sc, e1.y*ss+e2.y*sc, e1.z*ss+e2.z*sc};
    float3 v1=vc+e12;
	  v1=normalize(v1);
	  particles[i].en=gamma1;
    // second particle
	  float3 v2{-e1.x,-e1.y,-e1.z};
    if(particles[i].ip==3) v2-=vc; // Isotropic nA
	  else                   v2+=vc; // Forward EM
	  v2=normalize(v2);
    particles[i].vx=v1.x;
	  particles[i].vy=v1.y;
	  particles[i].vz=v1.z;
    particles[i+SHIFT].vx=v2.x;
    particles[i+SHIFT].vy=v2.y;
    particles[i+SHIFT].vz=v2.z;
    particles[i+SHIFT].en=gamma2;
    particles[i].ir=1.;
    particles[i+SHIFT].ir=1.;
  }
} // End of Reactor

void injector()
{
  static int push=0;
#ifdef OPENACC
  #pragma acc parallel loop gang vector present(Length,Begin,Counter)
#else
  #pragma omp parallel for
#endif
  for(int i=0; i<cub3; ++i)
  {
    Length[i]=0;
    Begin[i]=0;
    Counter[i]=0;
  }

//
#ifdef OPENACC
  #pragma acc parallel loop gang vector present(Length)
#else
  #pragma omp parallel for
#endif
  for(int i=0; i<LIFE; ++i)
  {
    const int vn=Get1DVoxelIndex(particles[i]);
    if(vn>-1 && vn<cub3)
    {
  #ifdef OPENACC
    #pragma acc atomic update
  #else
    #pragma omp atomic update
  #endif
      Length[vn]++;
    }
  }
#ifdef OPENACC
  #pragma acc serial present(Begin,Length)
  {
#endif
  Begin[0]=0;
  for(int i=1; i<cub3; ++i) Begin[i]=Begin[i-1]+Length[i-1];
#ifdef OPENACC
  }
#endif

#ifdef OPENACC
  #pragma acc parallel loop gang vector copy(LIFE,POSITION23,SHIFT) present(particles,Length,Counter,MIndex)
#else
  #pragma omp parallel for
#endif
  for(int vn=0; vn<cub3; ++vn)
  {
    if(Length[vn]>0)
    {
      //#pragma omp simd
      for(int i=0; i<POSITION23; ++i) // Collect all neutrons from reactions
      {
        if(particles[i].ip==3)
        {
          int ic=Get1DVoxelIndex(particles[i]);
          if(ic==vn)
          {
            MIndex[Begin[vn]+Counter[vn]]=i;
            //#pragma omp atomic update
            Counter[vn]++;
          }
        }
        //KMV told that everywhere where there is if(particles[i].ip==3)
        //we should add else{particles[i].ir=0}
        //i. e. if the particle is not a neutron, we should kill it.
        int j=i+SHIFT;
        if(particles[j].ip==3)
        {
          int ic=Get1DVoxelIndex(particles[j]);
          if(ic==vn)
          {
            MIndex[Begin[vn]+Counter[vn]]=j;
            //#pragma omp atomic update
            Counter[vn]++;
          }
        }
      }
      //#pragma omp simd
      for(int i=POSITION23; i<LIFE-POSITION23; ++i) // ir=1
      {
      
        if(particles[i].ix>=cuba || particles[i].jy>=cuba || particles[i].kz>=cuba || particles[i].ix<-cuba ||
           particles[i].jy<-cuba || particles[i].kz<-cuba)
          printf("**Comp1* i=%d j=%d k=%d ip=%d ir=%d\n",
                 particles[i].ix,particles[i].jy,particles[i].kz,
                 particles[i].ip,particles[i].ir);
        if(particles[i].ip==3)
        {
          int ic=Get1DVoxelIndex(particles[i]);
          if(ic==vn)
          {
            MIndex[Begin[vn]+Counter[vn]]=i;
            //#pragma omp atomic update
            Counter[vn]++;
          }
        }
      }
    }
  }
 
  for(int i=0; i<cub3; ++i) if(Counter[i] != Length[i]) printf("***ERROR: Counter[%d]=%d != Length[%d]=%d\n",i,Counter[i],i,Length[i]);

#ifdef OPENACC
  #pragma acc parallel loop gang vector present(particles,Length,Begin,MIndex,MMN)
#else
  #pragma omp parallel for schedule(dynamic)
  #pragma distribute_point
#endif
  for(int i=0; i<cub3; ++i)
  {
      int sn=Length[i];
      if(sn>1)
      {
        for(int k=0; k<sn; ++k)
        {
          int ind1=MIndex[Begin[i]+k];
          float3 p1{particles[ind1].vx,particles[ind1].vy,particles[ind1].vz};
          p1*=particles[ind1].en;
          bool brk=false;
#ifdef OPENACC
  #pragma acc loop vector
#else          
  #pragma omp simd
#endif        
          for(int n=0; n<k; ++n)
          {
            int ind2=MIndex[Begin[i]+n];
            if(particles[ind2].ir > 0)
            {
              float3 p2{particles[ind2].vx,particles[ind2].vy,particles[ind2].vz};
              p2*=particles[ind2].en;
              if(mod2(p2-p1) < fuse)
              {
                MMN[Begin[i]+n]=ind2;
                brk = true;
              }
              else MMN[Begin[i]+n]=-1;
            }
            else MMN[Begin[i]+n]=-1;
          }
          if(brk)
          {
            float w1=particles[ind1].wt;
            double sw=w1;
            float se=particles[ind1].en;
            float sx=particles[ind1].rx;
            float sy=particles[ind1].ry;
            float sz=particles[ind1].rz;
            float px=p1.x;
            float py=p1.y;
            float pz=p1.z;
#ifdef OPENACC
  #pragma acc loop vector reduction(+:sw,se,sx,sy,sz,px,py,pz)          
#else           
  #pragma omp simd reduction(+:sw,se,sx,sy,sz,px,py,pz)
#endif            
            for(int n=0; n<k; ++n)
            {
              if(MMN[Begin[i]+n]>-1)
              {
#ifdef DEB_NAN
                ++cnt;
#endif
                int ind2=MMN[Begin[i]+n];
                float w2=particles[ind2].wt;
                sw+=w2;                   
                float r21=w2/w1;
                sx+=particles[ind2].rx*r21;
                sy+=particles[ind2].ry*r21;
                sz+=particles[ind2].rz*r21;
                float e2=particles[ind2].en;
                se+=e2*r21;                
                float re21=r21*e2;
                px+=particles[ind2].vx*re21;
                py+=particles[ind2].vy*re21;
                pz+=particles[ind2].vz*re21;
                particles[ind2].ir=0;       
                particles[ind2].en=0.f;
                particles[ind2].de=0.f;
              }
            }
            double ps=sqrt((double)px*(double)px+(double)py*(double)py+(double)pz*(double)pz);
            particles[ind1].vx=px/ps;
            particles[ind1].vy=py/ps;
            particles[ind1].vz=pz/ps;
             double r1s=(double)w1/sw;
            particles[ind1].rx=sx*r1s; // new normalized coordinates
            particles[ind1].ry=sy*r1s;
            particles[ind1].rz=sz*r1s;
            
            particles[ind1].en=se*r1s; // new normalized energy
            particles[ind1].wt=sw;     // new weight
            
          } // End of if(brk)
        } // End of k loop
      } // End of if(sn1)>1
  }//End of 512 loop
  /*
  if(LIFE>GL)
  {
    std::cout<<"***Error: LIFE>GL "<<"LIFE="<<LIFE<<" GL="<<GL<<std::endl;
  }
  */

  NoNew=0.;
  double sg=0.;
  double wt=0.;
#ifdef OPENACC
  #pragma acc parallel loop gang vector reduction(+:sg,wt) present(particles)
#else  
  #pragma omp parallel for simd reduction(+:sg,wt)
#endif
  for(int j=0; j<LIFE; ++j)
  {
    sg+=(double)particles[j].en*(double)particles[j].wt;
    wt+=(double)particles[j].wt;
  }
  GAMMA=sg;
  NoNew+=wt;

  /*
  if(STEP==157)
  for(int i1=0; i1<1024; ++i1)
  {
    std::cout<<" i1="<<i1<<":";
    for(int i2=0; i2<32; ++i2)
    {
      int i=i1*32+i2;
      std::cout<<nnt[i].second<<" ";
    }
    std::cout<<std::endl;
  }
  */

  if(LIFE>Ntop) Ntop=LIFE;
  if(push<N)
  {
    if(LIFE<K) {make_particle(); ++push;}

    //std::cout<<"push particle n="<<push<<", M="<<Ntop<<"<"<<GLt<<", L="<<SLIFE
    //         <<", W="<<NoNew<<", G="<<GAMMA<<", DG="<<SumDGam<<", Tot="
    //         <<GAMMA+SumDGam<<std::endl;
  }
  else
  {
    //std::cout<<"fin o="<<over++<<", G="<<GAMMA<<", SDG="<<SumDGam<<", L="
    //         <<SLIFE<<"("<<SLIFE/oldL<<")"<<Ngn*Nth<<", W="<<NoNew<<"("<<NoNew/oldW
    //         <<"), Tot="<<GAMMA+SumDGam<<std::endl;
    //std::cout<<over++<<" "<<SLIFE<<" "<<NoNew<<std::endl;
    
    oldL=LIFE;
    oldW=NoNew;
  }
} // End of Injector

//////////////////////////////////////////////////////////////////////////////
// Program main
//////////////////////////////////////////////////////////////////////////////
int main(int argc, char **argv)
{
  //for(int i=0; i<cub3; ++i) Uburnout[i]=0.0;
	auto begin=std::chrono::steady_clock::now();
  LIFE=0;
  SumDGam=0.;
  MAX_ELEMENT=0;
  int step=1;
#ifdef OPENACC
#pragma acc data create(ind01,ind23,arr1,arr2,arr3) copyin(particles,Length,Begin,Counter,MIndex,MMN,Uburnout)
  {
#endif
    make_particle();
    while(step<max_loop && LIFE>0)
    {
      propagator();
      compressor();
      reactor();
      injector();     
      std::cout<<step<<"   "<<LIFE<<"    "<<NoNew<<"    "<<TimeLimit/DT
               <<"   "<<NumOf<<std::endl;
      STEP=step;
      ++step;
    }
#ifdef OPENACC
  }
#endif
	auto end=std::chrono::steady_clock::now();
	auto elapsed_ms=std::chrono::duration_cast<std::chrono::milliseconds>(end-begin);
	std::cout<<"time="<<elapsed_ms.count()<<" ms, G="<<G<<", K="<<K<<", Ntop="
           <<Ntop<<", SumDG="<<SumDGam<<std::endl;
  std::cout<<"Nbin="<<Nbin<<std::endl;
}
